import os
from unittest import defaultTestLoader, TestCase

import mocker
import simplejson

from lazr.js.meta import Builder, extract_metadata


class TestBuilder(Builder):
    """A special CSSComboFile that doesn't log to the stdout."""

    def log(self, msg):
        pass


class ExtractMetadataTest(TestCase):

    def test_extract_single_quote_raises_exception(self):
        """
        Extracting metadata fails if you have single quotes in your
        module registration block.
        """
        self.assertRaises(ValueError, extract_metadata, """\
        YUI.add('lazr.base', function(Y){
           Y.log('Hello World');
        }, '0.1', {'requires': ["node", "base"]});
        """)

    def test_extract_no_quote_raises_exception(self):
        """
        Not having quotes around your hash keys also causes an
        exception.
        """
        self.assertRaises(ValueError, extract_metadata, """\
        YUI.add('lazr.base', function(Y){
           Y.log('Hello World');
        }, '0.1', {requires: ["node", "base"]});
        """)

    def test_extract_single_module(self):
        """
        Extracting the metadata of a file containing a single module
        should successfully extract it's requirements.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.base', function(Y){
           Y.log('Hello World');
        }, '0.1', {"requires": ["node", "base"]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.base")
        self.assertEquals(metadata[0]["requires"], ["node", "base"])

    def test_extract_multiple_modules(self):
        """
        Extracting the metadata of a file that has more than one
        module registered should find all the registrations.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.base', function(Y){
           Y.log('Hello World');
        }, '0.1', {"requires": ["node", "base"]});
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, '0.1', {"requires": ["node", "anim", "event"]});
        """)

        self.assertEquals(len(metadata), 2)

        self.assertEquals(metadata[0]["name"], "lazr.base")
        self.assertEquals(metadata[0]["requires"], ["node", "base"])

        self.assertEquals(metadata[1]["name"], "lazr.anim")
        self.assertEquals(metadata[1]["requires"], ["node", "anim", "event"])

    def test_extract_multi_line(self):
        """
        If the module registration is split between more than one
        line, we should still be able to extract it.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, '0.1', {"requires": ["node", "anim",
                              "event"]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.anim")
        self.assertEquals(metadata[0]["requires"], ["node", "anim", "event"])

    def test_extract_odd_spacing(self):
        """
        Extracting metadata should work just fine if you have odd
        spacing between the requirements.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, '0.1', {"requires": [ "node" ,"anim"  ,  "event" ]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.anim")
        self.assertEquals(metadata[0]["requires"], ["node", "anim", "event"])

    def test_extract_with_use(self):
        """
        Having both 'use' and 'require in the same declaration should
        not break our parser.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, '0.1', {"use": ["dom"], "requires": [ "node" ,"anim"  ,  "event" ]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.anim")
        self.assertEquals(metadata[0]["requires"], ["node", "anim", "event"])
        self.assertEquals(metadata[0]["use"], ["dom"])

    def test_extract_with_requires_before_use(self):
        """
        The order between 'use' and 'requires' should not matter.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, '0.1', {"requires": [ "node" ,"anim"  ,  "event" ], "use": ["dom"]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.anim")
        self.assertEquals(metadata[0]["requires"], ["node", "anim", "event"])
        self.assertEquals(metadata[0]["use"], ["dom"])

    def test_extract_requires_in_new_line(self):
        """
        Even if the metadata is split between multiple lines, we
        should still be able to extract it.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, '0.1', {
        "requires": [
          "node" ,"anim"  ,  "event"
        ],
        "use": [
          "dom"
        ]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.anim")
        self.assertEquals(metadata[0]["requires"], ["node", "anim", "event"])
        self.assertEquals(metadata[0]["use"], ["dom"])

    def test_extract_with_use_and_no_requires(self):
        """
        No failure should happen if we have 'use' but no 'requires'.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, '0.1', {"use": ["dom"]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.anim")
        self.assertEquals(metadata[0]["use"], ["dom"])

    def test_extract_with_all_options(self):
        """
        A very complex registration, with all the options. We should
        be able to parse all of it.
        """
        metadata = extract_metadata("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
           X = Y.bind(function(Y){
               Y.log('goodbye world');
           }, Y);
        }, '0.1', {"use": ["dom"], "omit": ["nono"],
                   "optional": ["meh"],
                   "supersedes": ["old-anim"],
                   "after": ["lazr.base"]});
        """)

        self.assertEquals(len(metadata), 1)
        self.assertEquals(metadata[0]["name"], "lazr.anim")
        self.assertEquals(metadata[0]["use"], ["dom"])
        self.assertEquals(metadata[0]["omit"], ["nono"])
        self.assertEquals(metadata[0]["optional"], ["meh"])
        self.assertEquals(metadata[0]["supersedes"], ["old-anim"])
        self.assertEquals(metadata[0]["after"], ["lazr.base"])

class GenerateMetadataTest(mocker.MockerTestCase):

    def test_generate_metadata_simple(self):
        """
        Verify that the builder can extract and generates a metadata
        file in the expected format.
        """
        root = self.makeDir()
        anim = self.makeDir(path=os.path.join(root, "anim"))
        self.makeFile("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, "0.1", {"use": ["dom"]});
        """, basename="anim.js", dirname=anim)

        output = self.makeFile("")
        builder = TestBuilder(name="LAZR_CONFIG", src_dir=root,
                              output=output, exclude_regex="")
        builder.do_build()

        got = open(output, "r").read()
        prefix = got[:18]
        modules = got.splitlines()[-2].strip()[7:-1]
        self.assertEquals(prefix, "var LAZR_CONFIG = ")
        self.assertTrue('"path": "anim/anim-min.js"' in modules)
        self.assertFalse('"skinnable": false' in modules)
        self.assertTrue('"type": "js"' in modules)
        self.assertTrue('"use": ["dom"]' in modules)
        self.assertTrue('"ext": true' in modules)

    def test_extract_skinnable(self):
        """
        If a module is skinnable, we'll generate the module
        registration for it's skin ourselves, to prevent YUI from
        auto-generating one, since it assumes the skin path starts
        with the module name.
        """
        root = self.makeDir()
        anim = self.makeDir(path=os.path.join(root, "anim"))
        self.makeFile("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, "0.1", {"use": ["dom"], "requires": ["widget"], "skinnable": true});
        """, basename="anim.js", dirname=anim)

        skin = self.makeDir(path=os.path.join(anim, "assets", "skins", "sam"))
        self.makeFile("""\
        /* anim-skin.css */
        """, basename="anim-skin.css", dirname=skin)

        output = self.makeFile("")
        builder = TestBuilder(name="LAZR_CONFIG", src_dir=root,
                              output=output, exclude_regex="",
                              prefix="lazr")
        builder.do_build()

        got = open(output, "r").read()
        prefix = got[:18]
        modules = got.splitlines()[-2].strip()[7:-1]
        self.assertEquals(prefix, "var LAZR_CONFIG = ")
        self.assertTrue(
            '"path": "lazr/anim/assets/skins/sam/anim-skin.css"' in modules)
        self.assertTrue(
            '"path": "lazr/anim/anim-min.js"' in modules)
        self.assertTrue('"skinnable": true' in modules)
        self.assertTrue('"type": "js"' in modules)
        self.assertTrue('"type": "css"' in modules)
        self.assertTrue('"use": ["dom"]' in modules)
        self.assertTrue('"requires": ["widget"]' in modules)
        self.assertTrue(
            ('"after": ["cssreset", "cssfonts", "cssgrids", '
             '"cssreset-context", "cssfonts-context", "cssgrids-context", '
             '"skin-sam-widget"]') in modules)
        self.assertTrue('"ext": true' in modules)

    def test_extract_skinnable_with_lazr_conventions(self):
        """
        LazrJS conventionally uses a "-core.css" file to load core CSS rules,
        and a "-skin.css" file to store skinnable rules.

        If we find this filesystem structure, let's generate different
        CSS modules and hook them up accordingly. The filename doesn't
        need to match the module name.
        """
        root = self.makeDir()
        anim = self.makeDir(path=os.path.join(root, "anim"))
        self.makeFile("""\
        YUI.add('lazr.anim', function(Y){
           Y.log('Hello World');
        }, "0.1", {"use": ["dom"], "requires": ["widget"], "skinnable": true});
        """, basename="anim.js", dirname=anim)

        assets = self.makeDir(path=os.path.join(anim, "assets"))
        self.makeFile("""\
        /* purty-anim-core.css */
        """, basename="purty-anim-core.css", dirname=assets)

        skin = self.makeDir(path=os.path.join(assets, "skins", "sam"))
        self.makeFile("""\
        /* purty-anim-skin.css */
        """, basename="purty-anim-skin.css", dirname=skin)

        output = self.makeFile("")
        builder = TestBuilder(name="LAZR_CONFIG", src_dir=root,
                              output=output, exclude_regex="",
                              prefix="lazr")
        builder.do_build()

        got = open(output, "r").read()
        prefix = got[:18]
        lines = got.splitlines()
        modules = lines[-2].strip()[7:-1]
        variables = {}
        for line in lines[1:-2]:
            variable, literal = line.split("=")
            variable = variable.strip(" ").split(" ")[-1]
            variables[variable] = literal.strip(" ").split("\"")[1]
        self.assertEquals(prefix, "var LAZR_CONFIG = ")

        # Variables are only defined for optimizing minification if
        # they are used more than once, see how "dom" and "js" are not
        # included in this list, even though they are used by the
        # module.  All the other names are used more than once due to
        # the two CSS files that are part of the skin for this module.
        self.assertEquals(variables, {
            "CSS": "css",
            "CSSFONTS": "cssfonts",
            "CSSFONTS_CONTEXT": "cssfonts-context",
            "CSSGRIDS": "cssgrids",
            "CSSGRIDS_CONTEXT": "cssgrids-context",
            "CSSRESET": "cssreset",
            "CSSRESET_CONTEXT": "cssreset-context",
            "SKIN_SAM_LAZR_ANIM_CORE": "skin-sam-lazr.anim+core",
            "SKIN_SAM_WIDGET": "skin-sam-widget"})

        self.assertTrue(
            '"path": "lazr/anim/assets/purty-anim-core.css"' in modules)
        self.assertTrue(
            '"path": "lazr/anim/assets/skins/sam/purty-anim-skin.css"'
            in modules)
        self.assertTrue(
            '"path": "lazr/anim/anim-min.js"' in modules)
        self.assertTrue('"skinnable": true' in modules)
        self.assertTrue('"type": "js"' in modules)
        self.assertTrue('"type": CSS' in modules)
        self.assertTrue('"use": ["dom"]' in modules)
        self.assertTrue('"requires": ["widget"]' in modules)
        self.assertTrue('"requires": [SKIN_SAM_LAZR_ANIM_CORE]' in modules)
        self.assertTrue(
            ('"after": [CSSRESET, CSSFONTS, CSSGRIDS, '
            'CSSRESET_CONTEXT, CSSFONTS_CONTEXT, CSSGRIDS_CONTEXT, '
            'SKIN_SAM_WIDGET, SKIN_SAM_LAZR_ANIM_CORE]') in modules)
        self.assertTrue('"ext": true' in modules)


def test_suite():
    return defaultTestLoader.loadTestsFromName(__name__)
